/*
    Copyright  1995-2001, The AROS Development Team. All rights reserved.
    $Id: formatdateu.c,v 1.5 2023/03/16 19:14:31 jacadcaps Exp $

    Desc:
    Lang: English
*/

#include "locale_intern.h"
#include <exec/types.h>
#include <utility/hooks.h>
#include <utility/date.h>
#define __NOLIBBASE__
#include <proto/utility.h>
#undef __NOLIBBASE__
#include <clib/alib_protos.h>

#include "formatdate.h"

CONST_WSTRPTR GetLocaleStrU(struct Locale *, ULONG, struct LocaleBase *);

#define GetLocaleStrU(a,b) GetLocaleStrU(a, b, LocaleBase)

static VOID PrintDigits(UWORD number, LONG fill, UWORD len, struct Hook *hook, struct Locale *locale);
static VOID _WriteChar(ULONG token, struct Hook *hook, struct Locale *locale);
static VOID _WriteString(CONST_WSTRPTR string, struct Hook *hook, struct Locale *locale, LONG cmod);

static VOID _FormatDate(struct Locale *locale,
                 CONST_WSTRPTR formatString,
                 struct DateStamp *date,
                 struct ClockData *cData,
                 struct Hook *hook,
                 struct LocaleBase *LocaleBase, LONG def_fillchar, LONG def_cmod);

ULONG ToUpperU(ULONG);
ULONG ToLowerU(ULONG);

VOID FormatDateU(struct Locale *locale, CONST_WSTRPTR formatString, struct DateStamp *date, struct Hook *hook, struct LocaleBase *LocaleBase)
{
    struct ClockData cData;

    if (!hook)
	return;

    if (formatString)
    {
#warning Amiga2Date will fail around year 2114, because then the numer of seconds since 1978 dont fit in a 32 bit variable anymore!
   
	Amiga2Date(date->ds_Days * 86400 + date->ds_Minute * 60 + date->ds_Tick / 50,
	           &cData);

	_FormatDate(locale, formatString, date, &cData, hook, LocaleBase, DEF_FILLCHAR, DEF_CASEMOD);
    }

    _WriteChar('\0', hook, locale);	/* Write null terminator */
} /* FormatDate */


static VOID _FormatDate(struct Locale *locale,
                 CONST_WSTRPTR formatString,
                 struct DateStamp *date,
                 struct ClockData *cData,
                 struct Hook *hook,
                 struct LocaleBase *LocaleBase, LONG def_fillchar, LONG def_cmod)
{
    ULONG week, days, tmp;
    ULONG c;

    while ((c = *formatString++))
    {
    	if (c == '%')
		{
			LONG fillchar = def_fillchar, cmod = def_cmod;

	    c = *formatString++;

		switch (c)
		{
			case '_':
				fillchar = ' ';
				c = *formatString++;
				break;

			case '0':
				fillchar = '0';
				c = *formatString++;
				break;

			case '-':
				fillchar = FCHAR_NONE;
				c = *formatString++;
				break;

			case '^':
				cmod = CASEMOD_UPPER;
				c = *formatString++;
				break;
		}

	    if (!c)
		break;

	    switch (c)
	    {
	    case 'a':
		_WriteString(GetLocaleStrU(locale, ABDAY_1 + cData->wday), hook,
			     locale, cmod);
		break;
		
	    case 'A':
		_WriteString(GetLocaleStrU(locale, DAY_1 + cData->wday), hook,
			     locale, cmod);
		break;
		
	    case 'b':
	    case 'h':
		_WriteString(GetLocaleStrU(locale, ABMON_1 + cData->month - 1),
			    hook, locale, cmod);
		break;
		
	    case 'B':
		_WriteString(GetLocaleStrU(locale, MON_1 + cData->month - 1), hook,
			     locale, cmod);
		break;
		
	    case 'c':
		_FormatDate(locale, (CONST_WSTRPTR)L"%a %b %d %H:%M:%S %Y", date, cData, hook, LocaleBase, fillchar, cmod);
		break;
		
	    case 'C':
		_FormatDate(locale, (CONST_WSTRPTR)L"%a %b %e %T %Y", date, cData, hook, LocaleBase, fillchar, cmod);
		break;
		
	    case 'd':
		PrintDigits(cData->mday, FILLCHAR('0'), 2, hook, locale);
		break;
		
	    case 'x':
	    case 'D':
		_FormatDate(locale, (CONST_WSTRPTR)L"%m/%d/%y", date, cData, hook, LocaleBase, fillchar, cmod);
		break;
		
	    case 'e':
		PrintDigits(cData->mday, FILLCHAR(' '), 2, hook, locale);
		break;

	    case 'F':
		_FormatDate(locale, (CONST_WSTRPTR)L"%Y-%m-%d", date, cData, hook, LocaleBase, fillchar, cmod);
		break;

	    case 'H':
		PrintDigits(cData->hour, FILLCHAR('0'), 2, hook, locale);
		break;
		
	    case 'I':
		c = cData->hour;
		PrintDigits(c > 12 ? c - 12 : c == 0 ? 12 : c, FILLCHAR('0'), 2, hook, locale);
		break;
		
	    case 'j':
#if 1
		/* 15-Aug-2003 bugfix: Now works like the original. - Piru
		 */
		if (date)
		{
		    UWORD jul = date->ds_Days
		                - (cData->year * 365)
		                - (cData->year / 4)
		                + (cData->year / 100)
		                - (cData->year / 400) + 1;

		    PrintDigits(jul,
		                FILLCHAR('0'),
		                3,
		                hook,
		                locale);
		}
#else
	        /* TODO */
#warning Julian date not tested.
		/* Julian date is DDD (1 - 366)*/
		PrintDigits(
		    cData->mday + dayspermonth[cData->month],
		    FILLCHAR('0'),
		    3,
		    hook,
		    locale
		);
#endif
		break;
		
	    case 'm':
		PrintDigits(cData->month, FILLCHAR('0'), 2, hook, locale);
		break;
		
	    case 'M':
		PrintDigits(cData->min, FILLCHAR('0'), 2, hook, locale);
		break;
		
	    case 'n':
		_WriteChar('\n', hook, locale);
		break;
		
	    case 'p':
		_WriteString(GetLocaleStrU(locale,
					  cData->hour < 12 ? AM_STR : PM_STR),
					  hook, locale, cmod);
		break;

	    case 'P':
			_WriteString(GetLocaleStrU(locale, cData->hour < 12 ? AM_STR : PM_STR), hook, locale, CASEMOD_LOWER);
			break;

	    case 'q':
		PrintDigits(cData->hour, FILLCHAR(-1), 2, hook, locale);
		break;
		
	    case 'Q':
		c = cData->hour;
		PrintDigits(c > 12 ? c - 12 : c == 0 ? 12 : c, FILLCHAR(-1), 2, hook, locale);
		break;
		
	    case 'r':
#if 0
		_FormatDate(locale, "%I:%M:%S ", date, cData, hook, LocaleBase, fillchar, cmod);

		/* 15-Aug-2003 bugfix: Original is not using locale here, but
		 * built-in strings! However using locale makes much more
		 * sense here, so this "feature" is not reproduced here. - Piru
		 */
		_WriteString(cData->hour < 12 ? "AM" : "PM", hook, locale, cmod);
#else
		_FormatDate(locale, (CONST_WSTRPTR)L"%I:%M:%S %p", date, cData, hook, LocaleBase, fillchar, cmod);
#endif
		break;
		
	    case 'R':
		_FormatDate(locale, (CONST_WSTRPTR)L"%H:%M", date, cData, hook, LocaleBase, fillchar, cmod);
		break;
		
	    case 'S':
		PrintDigits(cData->sec, FILLCHAR('0'), 2, hook, locale);
		break;
		
	    case 't':
		_WriteChar('\t', hook, locale);
		break;
		
	    case 'X':
	    case 'T':
		_FormatDate(locale, (CONST_WSTRPTR)L"%H:%M:%S", date, cData, hook, LocaleBase, fillchar, cmod);
		break;
		
	    case 'W': /* week number, Monday first day of week */
	    case 'U': /* week number, Sunday first day of week */
	        days = cData->mday + dayspermonth[cData->month]; 
		
		/* leap year ? */
		if ((cData->year % 4) == 0 && cData->month > 2)
		{
		  /*
		  ** 1700, 1800, 1900, 2100, 2200 re not leap years.
		  ** 2000 is a leap year.
		  ** -> if a year is divisible by 100 but not by 400 then
		  ** it is not a leap year!
		  */
		  if ((cData->year % 100) == 0 && (cData->year % 400) != 0)
		    ;
		  else
		    days++;
		}
		
		/* 
		** If January 1st is a Monday then the first week
		** will start with a Sunday January 7th if Sunday is the first day of the week
		** but if Monday is the first day of the week then Jan 1st will also be the
		** first day of the first week.
		*/
		/* 
		** Go to Saturday = last day of week if Sunday is first day of week
                ** Go to Sunday   = last day of week if Monday is first day of week
		*/
		if (c == 'U')
		{
		  /* Sunday is first day of the week */
  		  tmp = days + (6 - cData->wday);
  		}
  		else
  		{
  		  /* Monday is first day of week */
  		  if (cData->wday != 0)
  		    tmp = days + (7 - cData->wday);
  		  else
  		    tmp = days;
		}

		if (tmp < 7)
		{
		  week = 0;
		}
		else
		{
		  /* cut off the few days that belong to week 0 */
		  tmp -= (tmp % 7);
		  /* Calculate the full amount of weeks */
		  week = tmp / 7;
		}
		
                PrintDigits(week, FILLCHAR('0'), 2, hook, locale);
	    break;
		
	    case 'w':
		PrintDigits(cData->wday, FILLCHAR(-1), 1, hook, locale);
		break;
		
	    case 'y':
		PrintDigits(cData->year % 100, FILLCHAR('0'), 2, hook, locale);
		break;
		
	    case 'Y':
		PrintDigits(cData->year, FILLCHAR('0'), 4, hook, locale);
		break;

            //case 'Z':
            //    /* current timezone Unimplemented in 3.1 */
            //    break;

	    default:
		_WriteChar(c, hook, locale);
		break;
	    }
	}
	else
	{
	    _WriteChar(c, hook, locale);
	}
    }
}


static VOID _WriteString(CONST_WSTRPTR string, struct Hook *hook, struct Locale *locale, LONG cmod)
{
	UBYTE c;

	while ((c = *string++))
	{
		if (cmod == CASEMOD_UPPER)
			c = ToUpperU(c);
		else if (cmod == CASEMOD_LOWER)
			c = ToLowerU(c);

		_WriteChar(c, hook, locale);
    }
}


static VOID _WriteChar(ULONG token, struct Hook *hook, struct Locale *locale)
{
     AROS_UFC3(VOID, hook->h_Entry,
       AROS_UFCA(struct Hook *,   hook,   A0),
       AROS_UFCA(struct Locale *, locale, A2),
       AROS_UFCA(ULONG,            token,  A1)
     );
}


static VOID PrintDigits(UWORD number, LONG fill, UWORD len, struct Hook *hook,
		 struct Locale *locale)
{
    ULONG  buf[7];
    ULONG *ptr = &buf[6];
    int   i   = 0;
    
    buf[6] = 0;
    
    /*
     * 6-mar-2003: Fixed formatting of "0:10 am", it showed up as ":10 am".
     * - Piru
     *
     * while(number && i < len)
     */
    while ((number || !i) && i < len)
    {
	*--ptr = number % 10 + '0';
	number /= 10;
	i++;
    }

    while (len - i > 0  && (char)-1 != fill)
    {
      len--;
      _WriteChar(fill, hook, locale);
    }

    _WriteString((CONST_WSTRPTR)ptr, hook, locale, CASEMOD_NONE);
}

VOID UCS4_iFormatClockData(CONST struct Locale *locale, CONST_WSTRPTR formatString, CONST struct ClockData *cData, CONST struct Hook *hook, struct LocaleBase *LocaleBase)
{
	if (hook)
	{
		if (formatString)
		{
			_FormatDate((struct Locale *)locale, formatString, NULL, (struct ClockData *)cData, (struct Hook *)hook, LocaleBase, DEF_FILLCHAR, DEF_CASEMOD);
		}

		_WriteChar('\0', (struct Hook *)hook, (struct Locale *)locale);	/* Write null terminator */
	}
}
